/*
* ae.c
* a library for evaluating mathematical expressions in C programs based on Lua
* Luiz Henrique de Figueiredo <lhf@tecgraf.puc-rio.br>
* 11 Aug 2018 16:44:34
* This code is hereby placed in the public domain and also under the MIT license
*/

#include <ctype.h>
#include <string.h>

#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"
#include "mycompat.h"

#undef  iscsym
#define iscsym(c)	(isalnum(c) || ((c) == '_'))

#include "ae.h"
#include "lmathx.c"

static lua_State* L=NULL;

void ae_open(void)
{
 if (L!=NULL) return;
 L=luaL_newstate();
 if (L==NULL) return;
 luaopen_mathx(L);
 lua_settop(L,0);
 lua_pushnil(L);			/* slot for error message */
}

void ae_close(void)
{
 if (L==NULL) return;
 lua_close(L);
 L=NULL;
}

void ae_set(const char* name, double value)
{
 if (L==NULL) return;
 lua_pushnumber(L,value);
 lua_setglobal(L,name);
}

const char* ae_error(void)
{
 if (L==NULL) return "ae not open"; else return lua_tostring(L,1);
}

#define FUNCTION	"function"
#define RETURN		"return "

static int badinput(const char* expression)
{
 const char* b=expression;
 const char* e;
 for (;;)
 {
  int p,q;
  b=strstr(b,FUNCTION); if (b==NULL) return 0;
  e=b+sizeof(FUNCTION)-2;
  p=(b==expression) ? 0 : b[-1];
  q=e[1];
  if (!iscsym(p) && !iscsym(q)) return 1;
  b=e+1;
 }
 return 1;
}

typedef struct LoadS { const char* text[3]; size_t size[3]; int i; } LoadS;

static const char* getS(lua_State* L, void* data, size_t* size)
{
 LoadS* S=(LoadS*) data;
 int i=S->i++;
 *size=S->size[i];
 (void)L;
 return S->text[i];
}

static int compile(const char* expression)
{
 LoadS S;
 if (badinput(expression)) return 1;
 S.text[0]=RETURN;	S.size[0]=sizeof(RETURN)-1;
 S.text[1]=expression;	S.size[1]=strlen(expression);
 S.text[2]=NULL;	S.size[2]=0;
 S.i=0;
 lua_pushliteral(L,"ae:1: syntax error");
 return lua_load(L,getS,&S,"=ae","text");
}

double ae_eval(const char* expression)
{
 double value;
 int error=0;
 if (L==NULL) return 0.0;
 lua_getglobal(L,expression);		/* is it cached? */
 switch (lua_type(L,-1))
 {
  case LUA_TNIL:			/* no: compile, cache, and call it */
   error=compile(expression); if (error) break;
   lua_pushvalue(L,-1);
   lua_setglobal(L,expression);
  case LUA_TFUNCTION:			/* yes: call it */
   if (!lua_iscfunction(L,-1)) error=lua_pcall(L,0,1,0);
 }
 if (error)
 {
  value=0.0;
  lua_pushstring(L,strchr(lua_tostring(L,-1),' ')+1);
 }
 else if (lua_type(L,-1)==LUA_TNUMBER)
 {
  value=lua_tonumber(L,-1);
  lua_pushnil(L);
 }
 else if (lua_type(L,-1)==LUA_TBOOLEAN)
 {
  value=lua_toboolean(L,-1);
  lua_pushnil(L);
 }
 else
 {
  value=0.0;
  lua_pushliteral(L,"not a number");
 }
 lua_replace(L,1);			/* save error message */
 lua_settop(L,1);
 return value;
}
